package dev.fitko.fitconnect.api.domain.sender;

import dev.fitko.fitconnect.api.domain.model.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.callback.Callback;
import dev.fitko.fitconnect.api.domain.model.metadata.AdditionalReferenceInfo;
import dev.fitko.fitconnect.api.domain.model.metadata.AuthenticationInformation;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.metadata.payment.PaymentInformation;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.ReplyChannel;
import dev.fitko.fitconnect.api.domain.sender.steps.unencrypted.BuildStep;
import dev.fitko.fitconnect.api.domain.sender.steps.unencrypted.DataStep;
import dev.fitko.fitconnect.api.domain.sender.steps.unencrypted.DestinationStep;
import dev.fitko.fitconnect.api.domain.sender.steps.unencrypted.OptionalPropertiesStep;
import dev.fitko.fitconnect.api.domain.sender.steps.unencrypted.ServiceTypeStep;
import lombok.Getter;

import java.net.URI;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import static dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType.APPLICATION_JSON;
import static dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType.APPLICATION_XML;

@Getter
public final class SendableSubmission {

    private final UUID destinationId;
    private final UUID caseId;
    private final String data;
    private final SubmissionSchema submissionSchema;
    private final List<Attachment> attachments;
    private final String serviceName;
    private final String serviceIdentifier;
    private final List<AuthenticationInformation> authenticationInformation;
    private final AdditionalReferenceInfo additionalReferenceInfo;
    private final PaymentInformation paymentInformation;
    private final ReplyChannel replyChannel;
    private final Callback callback;

    private SendableSubmission(final Builder builder) {
        destinationId = builder.getDestinationId();
        caseId = builder.getCaseId();
        data = builder.getData();
        submissionSchema = new SubmissionSchema(builder.getSchemaUri(), builder.getDataMimeType());
        attachments = Collections.unmodifiableList(builder.getAttachments());
        serviceName = builder.getServiceName();
        serviceIdentifier = builder.getServiceIdentifier();
        authenticationInformation = builder.getAuthenticationInformation();
        replyChannel = builder.getReplyChannel();
        paymentInformation = builder.getPaymentInformation();
        callback = builder.getCallback();
        additionalReferenceInfo = builder.getAdditionalReferenceInfo();
    }

    public static DestinationStep Builder() {
        return new Builder();
    }

    @Getter
    private static final class Builder implements DestinationStep,
            ServiceTypeStep,
            DataStep,
            OptionalPropertiesStep,
            BuildStep {

        private UUID destinationId;
        private UUID caseId;
        private final List<Attachment> attachments = new ArrayList<>();
        private String data;
        private URI schemaUri;
        private MimeType dataMimeType;
        private String serviceName;
        private String serviceIdentifier;
        private List<AuthenticationInformation> authenticationInformation;
        private PaymentInformation paymentInformation;
        private ReplyChannel replyChannel;
        private AdditionalReferenceInfo additionalReferenceInfo;
        private Callback callback;

        private Builder() {
        }

        @Override
        public ServiceTypeStep setDestination(final UUID destinationId) {
            this.destinationId = destinationId;
            return this;
        }

        @Override
        public OptionalPropertiesStep addAttachments(final List<Attachment> attachments) {
            if (attachments != null && !attachments.isEmpty()) {
                this.attachments.addAll(attachments);
            }
            return this;
        }

        @Override
        public OptionalPropertiesStep addAttachment(final Attachment attachment) {
            if (attachment != null) {
                attachments.add(attachment);
            }
            return this;
        }

        @Override
        public OptionalPropertiesStep setReplyChannel(final ReplyChannel replyChannel) {
            this.replyChannel = replyChannel;
            return this;
        }

        @Override
        public OptionalPropertiesStep setJsonData(final String jsonData, final URI schemaUri) {
            data = jsonData;
            dataMimeType = APPLICATION_JSON;
            this.schemaUri = schemaUri;
            return this;
        }


        @Override
        public OptionalPropertiesStep setXmlData(final String xmlData, final URI schemaUri) {
            data = xmlData;
            dataMimeType = APPLICATION_XML;
            this.schemaUri = schemaUri;
            return this;
        }

        @Override
        public DataStep setServiceType(final String serviceIdentifier, final String serviceName) {
            this.serviceIdentifier = serviceIdentifier;
            this.serviceName = serviceName;
            return this;
        }

        @Override
        public OptionalPropertiesStep setAuthenticationInformation(final List<AuthenticationInformation> authenticationInformation) {
            this.authenticationInformation = authenticationInformation;
            return this;
        }

        @Override
        public OptionalPropertiesStep setPaymentInformation(final PaymentInformation paymentInformation) {
            this.paymentInformation = paymentInformation;
            return this;
        }

        @Override
        public OptionalPropertiesStep setApplicationDate(final LocalDate applicationDate) {
            this.additionalReferenceInfo = AdditionalReferenceInfo.builder().withApplicationDate(applicationDate.toString()).build();
            return this;
        }

        @Override
        public OptionalPropertiesStep setCallback(final URI callbackUri, final String callbackSecret) {
            callback = new Callback(callbackUri, callbackSecret);
            return this;
        }

        @Override
        public OptionalPropertiesStep setCase(final UUID caseId) {
           this.caseId = caseId;
           return this;
        }

        @Override
        public SendableSubmission build() {
            return new SendableSubmission(this);
        }

    }
}
