package dev.fitko.fitconnect.client;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.Status;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataEncryptionIssue;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataSchemaViolation;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.CreatedSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.sender.SendableEncryptedSubmission;
import dev.fitko.fitconnect.api.domain.sender.SendableSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectSenderException;
import dev.fitko.fitconnect.api.exceptions.internal.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import dev.fitko.fitconnect.api.exceptions.internal.ValidationException;
import dev.fitko.fitconnect.client.sender.ReplyReceiver;
import dev.fitko.fitconnect.client.sender.SubmissionSender;
import dev.fitko.fitconnect.client.testutil.LogCaptor;
import dev.fitko.fitconnect.client.util.ValidDataGuard;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.file.Path;
import java.time.Instant;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.core.IsEqual.equalTo;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class SenderClientTest {

    private final FitConnectService senderMock;
    private final SenderClient senderClient;

    private static final LogCaptor logs = new LogCaptor();

    private final UUID destinationId = UUID.randomUUID();

    private final RSAKey publicKey = generateRsaKey(2048).toPublicJWK();

    @BeforeEach
    public void clearLogs() {
        logs.clearEvents();
    }

    public SenderClientTest() throws JOSEException {
        final ApplicationConfig config = new ApplicationConfig();
        senderMock = mock(FitConnectService.class);
        senderClient = new SenderClient(senderMock, new SubmissionSender(config, senderMock, new ValidDataGuard(senderMock)), new ReplyReceiver(config, senderMock));
        setupTestMocks(destinationId);
    }

    @Test
    void testSendSubmissionWithJsonData() {

        // Given
        final RSAKey publicKey = this.publicKey;

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(announcedSubmission.getSubmissionId());

        final SubmissionSchema jsonSchema = new SubmissionSchema();
        jsonSchema.setSchemaUri(URI.create("https://json.mimetype"));
        jsonSchema.setMimeType(MimeType.APPLICATION_JSON);
        final Destination destination = getDestination(destinationId, jsonSchema);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.announceSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.sendSubmission(any())).thenReturn(expectedSubmission);
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);
        when(senderMock.validateJsonFormat(any(), eq(URI.create("https://json.mimetype")))).thenReturn(ValidationResult.ok());
        when(senderMock.validateMetadataSchema(any())).thenReturn(ValidationResult.ok());


        // When
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://json.mimetype"))
                .addAttachment(Attachment.fromPath(Path.of("src/test/resources/attachment.txt"), "text/plain"))
                .setCallback(URI.create("https://example.callback.net"), "callbackSecretWithLengthGreater32")
                .build();

        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNotNull(sentSubmission);
        assertThat(sentSubmission.getSubmissionId(), equalTo(expectedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    void testSendSubmissionWithXmlData() {

        // Given
        final RSAKey publicKey = this.publicKey;

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(announcedSubmission.getSubmissionId());

        final SubmissionSchema xmlSchema = new SubmissionSchema();
        xmlSchema.setSchemaUri(URI.create("https://xml.mimetype"));
        xmlSchema.setMimeType(MimeType.APPLICATION_XML);
        final Destination destination = getDestination(destinationId, xmlSchema);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.announceSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.sendSubmission(any())).thenReturn(expectedSubmission);
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);
        when(senderMock.validateMetadataSchema(any())).thenReturn(ValidationResult.ok());
        when(senderMock.validateXmlFormat(any())).thenReturn(ValidationResult.ok());

        // When
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setXmlData("<xml><test>data</test></xml>", URI.create("https://xml.mimetype"))
                .addAttachment(Attachment.fromPath(Path.of("src/test/resources/attachment.txt"), "text/plain"))
                .setCallback(URI.create("https://example.callback.net"), "callbackSecretWithLengthGreater32")
                .build();

        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNotNull(sentSubmission);
        assertThat(sentSubmission.getSubmissionId(), equalTo(expectedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    void testMimeTypeDetection() {

        // Given
        final var destinationId = UUID.randomUUID();
        setupTestMocks(destinationId);

        // When
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "name")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .addAttachments(List.of(
                        Attachment.fromPath(Path.of("src/test/resources/text_attachment_with_no_filetype"), "text/plain"),
                        Attachment.fromPath(Path.of("src/test/resources/pdf_attachment_with_no_extension"), "application/pdf")))
                .build();

        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNotNull(sentSubmission);
    }

    @Test
    void testMissingDataMimeTypeSchema() {

        // Given
        final var destinationId = UUID.randomUUID();

        final var destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService));
        destination.setDestinationId(destinationId);

        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = this.publicKey;

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.announceSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);

        // When
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                .build();

        final var exception = assertThrows(FitConnectSenderException.class, () -> senderClient.send(submission));

        // Then
        assertThat(exception.getMessage(), equalTo("Combination of provided MIME type 'application/json' and schema URI " +
                "'https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json' is not allowed by the destination"));
    }

    @Test
    void testWithEmptyAttachments() {

        // Given
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .addAttachments(List.of())
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNotNull(sentSubmission);
        logs.assertContains("No attachments to upload");
    }

    @Test
    void testWithAttachmentFileNull() {

        // Given
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .addAttachment(null)
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNotNull(sentSubmission);
        logs.assertContains("No attachments to upload");
    }

    @Test
    void testWithAttachmentPathNotPresent() {

        // Given
        final var path = Path.of("/non/existing/path");

        // When
        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> Attachment.fromPath(path, "plain/text"));

        // Then
        assertThat(exception.getMessage(), containsString("Reading attachment from path '" + path + "' failed"));
    }

    @Test
    void testWithAttachmentInputStreamNotReadable() {

        // Given
        final var inputStream = new InputStream() {
            @Override
            public int read() throws IOException {
                throw new IOException("reading from stream failed");
            }
        };

        // When
        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> Attachment.fromInputStream(inputStream, "plain/text", "non-existing-file", "test"));

        // Then
        assertThat(exception.getMessage(), containsString("Attachment could not be read from input-stream"));
    }

    @Test
    void testEncryptionError() {

        // Given
        when(senderMock.encryptBytes(any(), any(), anyString())).thenThrow(new EncryptionException("Encryption failed"));

        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        // When
        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> senderClient.send(submission));

        // Then
        assertThat(exception.getMessage(), containsString("Encryption failed"));
    }

    @Test
    void testAnnouncingSubmissionFailed() {

        // Given
        when(senderMock.announceSubmission(any())).thenThrow(new RestApiException("Announcing submission failed"));

        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        // When
        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> senderClient.send(submission));

        // Then
        assertThat(exception.getMessage(), containsString("Announcing submission failed"));

    }

    @Test
    void testRestApiCallFailed() {

        // Given
        when(senderMock.announceSubmission(any())).thenThrow(new RestApiException("Announcing submission failed"));

        final SendableSubmission submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        // When
        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> senderClient.send(submission));

        // Then
        assertThat(exception.getMessage(), containsString("Announcing submission failed"));
    }

    @Test
    void testGettingEncryptionKeyFailed() {

        // Given

        when(senderMock.getEncryptionKeyForDestination(any())).thenThrow(new RestApiException("Getting encryption key failed"));

        // When
        final SendableSubmission submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> senderClient.send(submission));

        //Then
        assertThat(exception.getMessage(), containsString("Getting encryption key failed"));

    }

    @Test
    void testSendSubmissionWithEncryptedData() {

        // Given
        final JWECryptoService cryptoService = new JWECryptoService(new HashService());

        final SendableEncryptedSubmission submission = SendableEncryptedSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setEncryptedMetadata(cryptoService.encryptObject(publicKey, new Metadata(), MimeType.APPLICATION_JSON.value()))
                .setEncryptedData(cryptoService.encryptBytes(publicKey, "fachdaten".getBytes(), "text/plain"))
                .addEncryptedAttachment(UUID.randomUUID(), cryptoService.encryptBytes(publicKey, "attachmentData".getBytes(), "text/plain"))
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        assertNotNull(sentSubmission);
        logs.assertContains("SUCCESSFULLY HANDED IN SUBMISSION");
    }

    @Test
    void testPublicEncryptionKeyRetrieval() {

        // Given
        final var destinationId = UUID.randomUUID();
        final var destination = getDestination(destinationId);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);

        // When
        final String encryptionKey = senderClient.getPublicKeyForDestination(destinationId);

        assertNotNull(encryptionKey);
        assertThat(encryptionKey, equalTo(publicKey.toJSONString()));
    }

    @Test
    void testFailOnInvalidMetadata() {

        // Given
        when(senderMock.validateMetadataSchema(any())).thenReturn(ValidationResult.error(new ValidationException("invalid metadata")));

        final SendableSubmission submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        // When
        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> senderClient.send(submission));

        //Then
        assertThat(exception.getCause().getMessage(), containsString("invalid metadata"));
    }

    @Test
    void testGetSubmissionStatus() {

        // Given
        final SentSubmission sentSubmission = SentSubmission.builder()
                .caseId(UUID.randomUUID())
                .submissionId(UUID.randomUUID())
                .destinationId(UUID.randomUUID())
                .authenticationTags(new AuthenticationTags())
                .build();

        final Status status = new Status(Event.ACCEPT_SUBMISSION.getState(), Date.from(Instant.now()), Collections.emptyList());

        when(senderMock.getStatus(sentSubmission)).thenReturn(status);

        // When
        final Status statusForSubmission = senderClient.getSubmissionStatus(sentSubmission);

        // Then
        assertThat(statusForSubmission.getState(), is(Event.ACCEPT_SUBMISSION.getState()));

    }

    @Test
    void testGetSubmissionStatusWithProblems() {

        // Given
        final SentSubmission sentSubmission = SentSubmission.builder()
                .caseId(UUID.randomUUID())
                .submissionId(UUID.randomUUID())
                .destinationId(UUID.randomUUID())
                .authenticationTags(new AuthenticationTags())
                .build();

        final Status status = new Status(Event.REJECT_SUBMISSION.getState(), Date.from(Instant.now()), List.of(new DataEncryptionIssue(), new DataSchemaViolation()));

        when(senderMock.getStatus(sentSubmission)).thenReturn(status);

        // When
        final Status statusForSubmission = senderClient.getSubmissionStatus(sentSubmission);

        // Then
        assertThat(statusForSubmission.getState(), is(Event.REJECT_SUBMISSION.getState()));
        assertThat(statusForSubmission.getProblems(), is(List.of(new DataEncryptionIssue(), new DataSchemaViolation())));

    }

    @Test
    void validateCallback() {

        when(senderMock.validateCallback(anyString(), anyLong(), anyString(), anyString())).thenReturn(ValidationResult.ok());

        final ValidationResult validationResult = senderClient.validateCallback("hmac", 0L, "body", "secret");

        verify(senderMock, times(1)).validateCallback(anyString(), anyLong(), anyString(), anyString());
        assertTrue(validationResult.isValid());
        assertFalse(validationResult.hasError());
    }

    private void setupTestMocks(final UUID destinationId) {

        final var destination = getDestination(destinationId);
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = this.publicKey;

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.announceSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);
        when(senderMock.validateMetadataSchema(any())).thenReturn(ValidationResult.ok());
        when(senderMock.validateXmlFormat(any())).thenReturn(ValidationResult.ok());
        when(senderMock.validateJsonFormat(any(), eq(URI.create("https://mimetype.test.de")))).thenReturn(ValidationResult.ok());
        when(senderMock.sendSubmission(any())).thenReturn(new Submission());

    }

    private CreatedSubmission getAnnouncedSubmission(final UUID destinationId) {
        final SubmissionSchema submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://json-schema.fitco.de"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("de:leika:test:key");
        destinationService.setSubmissionSchemas(Set.of(submissionSchema));

        final var announcedSubmission = new CreatedSubmission();
        announcedSubmission.setSubmissionId(UUID.randomUUID());
        announcedSubmission.setDestinationId(destinationId);
        return announcedSubmission;
    }

    private Destination getDestination(final UUID destinationId) {
        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://mimetype.test.de"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final var destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    private Destination getDestination(final UUID destinationId, final SubmissionSchema schema) {
        final var destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    private static RSAKey generateRsaKey(final int size) throws JOSEException {
        return new RSAKeyGenerator(size)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();
    }
}