package dev.fitko.fitconnect.core.http;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.config.http.HttpConfig;
import dev.fitko.fitconnect.api.config.http.ProxyConfig;
import dev.fitko.fitconnect.api.config.http.Timeouts;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import okhttp3.Headers;
import okhttp3.Interceptor;
import okhttp3.MediaType;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.RequestBody;
import okhttp3.Response;
import okhttp3.ResponseBody;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.Proxy;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.Collections;
import java.util.List;
import java.util.Map;

public class HttpClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(HttpClient.class);
    private static final ObjectMapper MAPPER = new ObjectMapper().configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    private final OkHttpClient httpClient;

    public HttpClient() {
        this(new HttpConfig(), Collections.emptyList());
    }

    public HttpClient(HttpConfig httpConfig, List<Interceptor> interceptors) {
        var builder = new OkHttpClient.Builder();
        setInterceptors(interceptors, builder);
        setTimeouts(httpConfig.getTimeouts(), builder);
        setProxy(httpConfig.getProxyConfig(), builder);
        httpClient = builder.build();
    }

    private static void setProxy(ProxyConfig proxyConfig, OkHttpClient.Builder builder) {
        if (proxyConfig.isProxySet()) {
            final Proxy proxy = proxyConfig.getHttpProxy();
            LOGGER.info("Creating HttpClient with proxy configuration: {}", proxy);
            builder.proxy(proxy);
        } else {
            LOGGER.info("Creating HttpClient without proxy configuration.");
        }
    }

    private static void setInterceptors(List<Interceptor> interceptors, OkHttpClient.Builder builder) {
        interceptors.forEach(builder::addInterceptor);
    }

    private static void setTimeouts(Timeouts timeouts, OkHttpClient.Builder builder) {
        builder.readTimeout(timeouts.getReadTimeoutInSeconds());
        builder.writeTimeout(timeouts.getWriteTimeoutInSeconds());
        builder.connectTimeout(timeouts.getConnectionTimeoutInSeconds());
    }

    public <R> HttpResponse<R> get(final String url, final Class<R> responseType) {
        return get(url, Map.of(), responseType);
    }

    public <R> HttpResponse<R> get(final String url, final Map<String, String> headers, final Class<R> responseType) {

        final Request request = new Request.Builder()
                .url(url)
                .get()
                .headers(Headers.of(headers))
                .build();

        try (final Response response = httpClient.newCall(request).execute()) {
            return evaluateStatusAndRespond(response, responseType);
        } catch (final IOException exception) {
            throw new RestApiException("HTTP GET call to '" + url + "' failed.", exception);
        }
    }

    public synchronized <P, R> HttpResponse<R> post(final String url, final Map<String, String> headers, final P httpPayload, final Class<R> responseType) {

        try {

            final RequestBody requestBody = createRequestBody(headers, httpPayload);
            final Request request = new Request.Builder()
                    .url(url)
                    .post(requestBody)
                    .headers(Headers.of(headers))
                    .build();

            try (final Response response = httpClient.newCall(request).execute()) {
                return evaluateStatusAndRespond(response, responseType);
            }

        } catch (final IOException exception) {
            throw new RestApiException("HTTP POST call to '" + url + "' failed.", exception);
        }
    }

    public synchronized <P, R> HttpResponse<R> put(final String url, final Map<String, String> headers, final P httpPayload, final Class<R> responseType) {

        try {

            final RequestBody requestBody = createRequestBody(headers, httpPayload);
            final Request request = new Request.Builder()
                    .url(url)
                    .put(requestBody)
                    .headers(Headers.of(headers))
                    .build();

            try (final Response response = httpClient.newCall(request).execute()) {
                return evaluateStatusAndRespond(response, responseType);
            }

        } catch (final IOException exception) {
            throw new RestApiException("HTTP PUT call to '" + url + "' failed.", exception);
        }
    }

    /**
     * The methods {@link ObjectMapper#writeValueAsBytes(Object)} and
     * {@link ObjectMapper#writeValueAsString(Object)} seem to have issues with writing
     * a raw string into a new string, resulting in obsolete quotation marks at begin and end of the actual string.
     * Therefore, this method contains a dedicated logic part for dealing with raw strings.
     * <p>
     * Also, we have to do a little workaround with {@link String#getBytes(Charset)} due to OkHttp enforcing a standard
     * value for the header "Content-Type" in case of raw strings. See
     * <a href="https://github.com/square/okhttp/issues/2099">https://github.com/square/okhttp/issues/2099</a>
     * for more information on that issue.
     */
    <T> RequestBody createRequestBody(final Map<String, String> headers, final T httpPayload) throws JsonProcessingException {

        final byte[] body;
        if (httpPayload.getClass().equals(String.class)) {
            body = ((String) httpPayload).getBytes(StandardCharsets.UTF_8);
        } else {
            body = MAPPER.writeValueAsBytes(httpPayload);
        }

        if (headers.containsKey(HttpHeaders.CONTENT_TYPE)) {
            return RequestBody.create(body, MediaType.parse(headers.get(HttpHeaders.CONTENT_TYPE)));
        }

        return RequestBody.create(body, MediaType.get(MimeTypes.APPLICATION_JSON));
    }

    <R> HttpResponse<R> evaluateStatusAndRespond(final Response response, final Class<R> responseType) throws IOException {

        if (!response.isSuccessful()) {
            final String message = response.body() != null ? response.body().string() : "HTTP call failed.";
            throw new RestApiException(message, response.code());
        }
        final ResponseBody responseBody = response.body();
        if (responseBody == null) {
            throw new RestApiException("Response body is null.");
        }
        if (responseType.equals(Void.class)) {
            return new HttpResponse<>(response.code(), null);
        }
        if (responseType.equals(String.class)) {
            return new HttpResponse<>(response.code(), (R) responseBody.string());
        }

        return new HttpResponse<>(response.code(), MAPPER.readValue(responseBody.string(), responseType));
    }
}
