package dev.fitko.fitconnect.core;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.domain.model.cases.Case;
import dev.fitko.fitconnect.api.domain.model.cases.Cases;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.Status;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataHashMismatch;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.MetadataSchemaViolation;
import dev.fitko.fitconnect.api.domain.model.event.problems.other.TechnicalError;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.AttachmentsMismatch;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentForValidation;
import dev.fitko.fitconnect.api.domain.model.reply.AcceptReply;
import dev.fitko.fitconnect.api.domain.model.reply.AnnounceReply;
import dev.fitko.fitconnect.api.domain.model.reply.CreatedReply;
import dev.fitko.fitconnect.api.domain.model.reply.RepliesForPickup;
import dev.fitko.fitconnect.api.domain.model.reply.Reply;
import dev.fitko.fitconnect.api.domain.model.reply.ReplyForPickup;
import dev.fitko.fitconnect.api.domain.model.reply.SentReply;
import dev.fitko.fitconnect.api.domain.model.reply.SubmitReply;
import dev.fitko.fitconnect.api.domain.model.submission.AnnounceSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.CreatedSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionsForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.internal.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.internal.EventLogException;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import dev.fitko.fitconnect.api.exceptions.internal.ValidationException;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.events.CaseService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.keys.KeyService;
import dev.fitko.fitconnect.api.services.reply.ReplyService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.net.URI;
import java.text.ParseException;
import java.time.Instant;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import static dev.fitko.fitconnect.api.domain.model.event.EventState.ACCEPTED;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.equalTo;
import static org.hamcrest.Matchers.instanceOf;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.ArgumentMatchers.anyList;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class FitConnectDefaultServiceTest {

    private final SubmissionService submissionServiceMock;
    private final CaseService caseServiceMock;
    private final ReplyService replyServiceMock;
    private final CryptoService cryptoServiceMock;
    private final ValidationService validationServiceMock;
    private final KeyService keyServiceMock;
    private final SecurityEventService securityEventServiceMock;
    private final FitConnectService underTest;

    public FitConnectDefaultServiceTest() {
        submissionServiceMock = mock(SubmissionService.class);
        caseServiceMock = mock(CaseService.class);
        replyServiceMock = mock(ReplyService.class);
        cryptoServiceMock = mock(CryptoService.class);
        validationServiceMock = mock(ValidationService.class);
        keyServiceMock = mock(KeyService.class);
        securityEventServiceMock = mock(SecurityEventService.class);

        underTest = new FitConnectDefaultService(
                submissionServiceMock,
                caseServiceMock,
                replyServiceMock,
                cryptoServiceMock,
                validationServiceMock,
                keyServiceMock,
                securityEventServiceMock
        );
    }

    @Nested
    class SenderRelatedTests {
        @Test
        void encryptBytes() throws JOSEException {

            //Given
            final RSAKey publicKey = new RSAKeyGenerator(2048)
                    .keyUse(KeyUse.ENCRYPTION)
                    .keyID(UUID.randomUUID().toString())
                    .generate()
                    .toPublicJWK();

            when(cryptoServiceMock.encryptBytes(any(), any(), any())).thenReturn("encryptedString");

            // When
            final String encryptedData = underTest.encryptBytes(publicKey, "data".getBytes(), "text/plain");

            // Then
            assertNotNull(encryptedData);
            assertThat(encryptedData, is("encryptedString"));
        }

        @Test
        void createHash() {

            // Given
            when(cryptoServiceMock.hashBytes(any())).thenReturn("hashedData");

            // When
            final String hashedData = underTest.createHash("data".getBytes());

            // Then
            assertNotNull(hashedData);
            assertThat(hashedData, is("hashedData"));
        }

        @Test
        void createSubmission() {

            // Given
            final var createSubmission = AnnounceSubmission.builder()
                    .caseId(UUID.randomUUID())
                    .destinationId(UUID.randomUUID())
                    .announcedAttachments(Collections.emptyList())
                    .serviceType(new ServiceType())
                    .build();

            final var expectedSubmission = new CreatedSubmission();
            expectedSubmission.setSubmissionId(UUID.randomUUID());
            expectedSubmission.setDestinationId(createSubmission.getDestinationId());
            expectedSubmission.setCaseId(createSubmission.getCaseId());

            when(submissionServiceMock.announceSubmission(any())).thenReturn(expectedSubmission);

            // When
            final CreatedSubmission submissionForPickup = underTest.announceSubmission(createSubmission);

            // Then
            assertThat(submissionForPickup, equalTo(expectedSubmission));
        }

        @Test
        void uploadSubmissionAttachment() {

            // Given
            final var submissionId = UUID.randomUUID();
            final var attachmentId = UUID.randomUUID();
            final var encryptedAttachment = "encryptedAttachmentBytes";

            doNothing().when(submissionServiceMock).uploadAttachment(any(), any(), any());

            // When
            underTest.uploadSubmissionAttachment(submissionId, attachmentId, encryptedAttachment);

            // Then
            verify(submissionServiceMock, times(1)).uploadAttachment(submissionId, attachmentId, encryptedAttachment);
        }

        @Test
        void getEncryptionKeyForDestination() throws ParseException, IOException {

            // Given
            final var submissionId = UUID.randomUUID();

            final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");
            final RSAKey expectedRSAKey = RSAKey.parse(publicKeyJson);

            when(keyServiceMock.getPublicEncryptionKey(any())).thenReturn(expectedRSAKey);

            // When
            final RSAKey key = underTest.getEncryptionKeyForDestination(submissionId);

            // Then
            assertThat(key, equalTo(expectedRSAKey));
        }

        @Test
        void getEncryptionKeyForDestinationFailed() {

            // Given
            when(keyServiceMock.getPublicEncryptionKey(any())).thenThrow(RestApiException.class);

            // Then
            assertThrows(
                    RestApiException.class,
                    () -> underTest.getEncryptionKeyForDestination(UUID.randomUUID()),
                    "Expected getEncryptionKeyForDestination() to throw, but nothing was thrown"
            );
        }

        @Test
        void encryptObject() throws JOSEException {

            //Given
            final RSAKey publicKey = new RSAKeyGenerator(2048)
                    .keyUse(KeyUse.ENCRYPTION)
                    .keyID(UUID.randomUUID().toString())
                    .generate()
                    .toPublicJWK();

            when(cryptoServiceMock.encryptObject(any(), any(), any())).thenReturn("encryptedObject");

            final Metadata metadata = new Metadata();

            // When
            final String encryptedMetadata = underTest.encryptObject(publicKey, metadata, "application/json");

            // Then
            assertThat(encryptedMetadata, equalTo("encryptedObject"));
        }

        @Test
        void encryptingObjectFailed() throws JOSEException {

            //Given
            final RSAKey publicKey = new RSAKeyGenerator(2048)
                    .keyUse(KeyUse.ENCRYPTION)
                    .keyID(UUID.randomUUID().toString())
                    .generate()
                    .toPublicJWK();

            // When
            when(cryptoServiceMock.encryptObject(any(), any(), any())).thenThrow(EncryptionException.class);

            // Then
            assertThrows(
                    EncryptionException.class,
                    () -> underTest.encryptObject(publicKey, new Metadata(), "application/json"),
                    "Expected encryptObject() to throw, but nothing was thrown"
            );
        }

        @Test
        void sendSubmission() {

            // Given
            final SubmitSubmission submitSubmission = new SubmitSubmission();
            submitSubmission.setSubmissionId(UUID.randomUUID());
            submitSubmission.setEncryptedData("encryptedData");
            submitSubmission.setEncryptedMetadata("encryptedMetadata");

            final Submission expectedSubmission = new Submission();
            expectedSubmission.setSubmissionId(submitSubmission.getSubmissionId());
            expectedSubmission.setEncryptedData(submitSubmission.getEncryptedData());
            expectedSubmission.setEncryptedMetadata(submitSubmission.getEncryptedMetadata());

            when(submissionServiceMock.sendSubmission(any())).thenReturn(expectedSubmission);

            // When
            underTest.sendSubmission(submitSubmission);

            // Then
            verify(submissionServiceMock, times(1)).sendSubmission(submitSubmission);
        }

        @Test
        void getDestination() {

            // Given
            final var destinationId = UUID.randomUUID();

            final var expectedDestination = new Destination();
            expectedDestination.setDestinationId(destinationId);

            when(submissionServiceMock.getDestination(any())).thenReturn(expectedDestination);

            // When
            final Destination destination = underTest.getDestination(destinationId);

            // Then
            assertThat(destination, equalTo(expectedDestination));
        }

        @Test
        void testValidMetadataSchema() {

            // Given
            final var metadata = new Metadata();

            when(validationServiceMock.validateMetadataSchema(any())).thenReturn(ValidationResult.ok());

            // When
            final ValidationResult validatedMetadata = underTest.validateMetadataSchema(metadata);

            // Then
            assertTrue(validatedMetadata.isValid());
        }

        @Test
        void testInvalidMetadataSchema() {

            // Given
            final var invalidMetadata = new Metadata();
            final ValidationResult expectedResult = ValidationResult.error(new ValidationException("invalid metadata"));
            when(validationServiceMock.validateMetadataSchema(any())).thenReturn(expectedResult);

            // When
            final ValidationResult validatedMetadata = underTest.validateMetadataSchema(invalidMetadata);

            // Then
            assertTrue(validatedMetadata.hasError());
            assertThat(validatedMetadata.getError().getMessage(), equalTo("invalid metadata"));
        }

        @Test
        void testValidJsonFormatValidation() {

            // Given
            when(validationServiceMock.validateSubmissionDataSchema(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            // When
            final ValidationResult validationResult = underTest.validateJsonFormat("{}", URI.create("urn:something"));

            //Then
            assertTrue(validationResult.isValid());
        }

        @Test
        void testInvalidJsonFormatValidation() {

            // Given
            when(validationServiceMock.validateSubmissionDataSchema(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.error(new ValidationException("Invalid json")));

            // When
            final ValidationResult validationResult = underTest.validateJsonFormat("$%&/()", URI.create("urn:something"));

            //Then
            assertTrue(validationResult.hasError());
            assertThat(validationResult.getError().getMessage(), containsString("Invalid json"));
        }

        @Test
        void testValidXmlFormatValidation() {

            // Given
            when(validationServiceMock.validateXmlFormat(any())).thenReturn(ValidationResult.ok());

            // When
            final ValidationResult validationResult = underTest.validateXmlFormat("<xml>test</xml>");

            //Then
            assertTrue(validationResult.isValid());
        }

        @Test
        void testInvalidXmlFormatValidation() {

            // Given
            when(validationServiceMock.validateXmlFormat(any())).thenReturn(ValidationResult.error(new ValidationException("Invalid xml")));

            // When
            final ValidationResult validationResult = underTest.validateXmlFormat("$%&/(");

            //Then
            assertTrue(validationResult.hasError());
            assertThat(validationResult.getError().getMessage(), containsString("Invalid xml"));
        }

        @Test
        void testReceiveValidEventLog() {

            // Given
            final var caseId = UUID.randomUUID();
            final var destinationId = UUID.randomUUID();

            final EventLogEntry eventLogEntry = EventLogEntry.builder()
                    .caseId(caseId)
                    .build();

            when(caseServiceMock.getEventLog(any(), any())).thenReturn(List.of(eventLogEntry));

            // When
            final List<EventLogEntry> eventLog = underTest.getEventLog(destinationId, caseId);

            // Then
            assertFalse(eventLog.isEmpty());
            assertThat(eventLog.get(0).getCaseId(), is(caseId));
        }

        @Test
        void testReceiveLatestStatusForSubmission() {

            // Given
            final UUID caseId = UUID.randomUUID();
            final UUID destinationId = UUID.randomUUID();
            final UUID submissionId = UUID.randomUUID();

            final Status expectedStatus = new Status(ACCEPTED, Date.from(Instant.now()), Collections.emptyList());
            final SentSubmission sentSubmission = new SentSubmission(destinationId, caseId, submissionId, new AuthenticationTags());

            when(caseServiceMock.getStatus(sentSubmission)).thenReturn(expectedStatus);

            // When
            final Status status = underTest.getStatus(sentSubmission);

            // Then
            assertThat(status.getState(), is(ACCEPTED));
        }
    }

    @Nested
    class SubscriberRelatedTests {

        @Test
        void decryptStringContent() throws JOSEException {

            // Given
            final RSAKey privateKey = new RSAKeyGenerator(2048)
                    .keyUse(KeyUse.ENCRYPTION)
                    .keyID(UUID.randomUUID().toString())
                    .generate();

            final var expectedDecryptedString = "decryptedString";

            when(cryptoServiceMock.decryptToBytes(any(), any())).thenReturn(expectedDecryptedString.getBytes());

            // When
            final byte[] decryptedBytes = underTest.decryptString(privateKey, "encryptedString");

            // Then
            assertThat(new String(decryptedBytes), equalTo(expectedDecryptedString));
        }

        @Test
        void pollAvailableSubmissions() {

            // Given
            final var destinationId = UUID.randomUUID();

            final SubmissionForPickup submission = new SubmissionForPickup();
            final SubmissionsForPickup availableSubmissions = new SubmissionsForPickup();
            availableSubmissions.setSubmissions(Set.of(submission));
            availableSubmissions.setCount(1);

            when(submissionServiceMock.pollAvailableSubmissionsForDestination(any(), anyInt(), anyInt())).thenReturn(availableSubmissions);

            // When
            final SubmissionsForPickup submissionsForPickups = underTest.getAvailableSubmissions(destinationId, 0, 100);

            // Then
            assertThat(submissionsForPickups.getSubmissions(), equalTo(availableSubmissions.getSubmissions()));
            assertThat(submissionsForPickups.getCount(), is(availableSubmissions.getSubmissions().size()));
        }

        @Test
        void getSubmission() {

            // Given
            final var submissionId = UUID.randomUUID();

            final var expectedSubmission = new Submission();
            expectedSubmission.setSubmissionId(submissionId);

            when(submissionServiceMock.getSubmission(any())).thenReturn(expectedSubmission);

            // When
            final Submission submission = underTest.getSubmission(submissionId);

            // Then
            assertNotNull(submission);
            assertThat(submission.getSubmissionId(), is(submissionId));
        }

        @Test
        void submissionNotFound() {

            // Given
            when(submissionServiceMock.getSubmission(any())).thenThrow(RestApiException.class);

            // Then
            assertThrows(
                    RestApiException.class,
                    () -> underTest.getSubmission(UUID.randomUUID()),
                    "Expected getSubmission() to throw, but nothing was thrown"
            );
        }

        @Test
        void testFetchAttachment() {

            // Given
            final var submissionId = UUID.randomUUID();
            final var attachmentId = UUID.randomUUID();
            final var expectedAttachment = "encryptedAttachmentContent";

            when(submissionServiceMock.getAttachment(any(), any())).thenReturn(expectedAttachment);

            // When
            final String encryptedAttachment = underTest.getSubmissionAttachment(submissionId, attachmentId);

            // Then
            assertThat(encryptedAttachment, equalTo(expectedAttachment));
        }

        @Test
        void testValidSubmissionMetadata() {

            // Given
            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());

            final Submission submission = new Submission();
            submission.setDestinationId(destination.getDestinationId());

            final Metadata metadata = new Metadata();
            final AuthenticationTags authenticationTags = new AuthenticationTags();

            when(submissionServiceMock.getDestination(destination.getDestinationId())).thenReturn(destination);
            when(validationServiceMock.validateSubmissionMetadata(metadata, submission, destination, authenticationTags)).thenReturn(ValidationResult.ok());

            // When
            final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, authenticationTags);

            // Then
            assertTrue(validationResult.isValid());
        }

        @Test
        void testInvalidSubmissionMetadata() {

            // Given
            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());

            final Submission submission = new Submission();
            submission.setDestinationId(destination.getDestinationId());

            final Metadata metadata = new Metadata();
            final AuthenticationTags authenticationTags = new AuthenticationTags();

            when(submissionServiceMock.getDestination(destination.getDestinationId())).thenReturn(destination);
            when(validationServiceMock.validateSubmissionMetadata(metadata, submission, destination, authenticationTags)).thenReturn(ValidationResult.problem(new MetadataSchemaViolation()));

            // When
            final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, authenticationTags);

            // Then
            assertTrue(validationResult.hasProblems());
            assertThat(validationResult.getProblems().get(0), instanceOf(MetadataSchemaViolation.class));
        }

        @Test
        void invalidData() {

            // Given
            when(validationServiceMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.problem(new DataHashMismatch()));

            // When
            final ValidationResult validationResult = underTest.validateData("foo".getBytes(), "§$%&/(", new Metadata(), "");

            // Then
            assertTrue(validationResult.hasProblems());
            assertThat(validationResult.getProblems().get(0), instanceOf(DataHashMismatch.class));
        }

        @Test
        void testAcceptSubmission() throws ParseException, IOException {

            // Given
            final var submission = new Submission();
            submission.setSubmissionId(UUID.randomUUID());
            submission.setDestinationId(UUID.randomUUID());
            submission.setCaseId(UUID.randomUUID());

            final EventPayload eventPayload = EventPayload.forAcceptEvent(submission);

            final var eventJWT = getResourceAsString("/signed_jwt.txt");

            when(securityEventServiceMock.createAcceptSubmissionEvent(any())).thenReturn(SignedJWT.parse(eventJWT));
            when(validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
            doNothing().when(caseServiceMock).sendEvent(any(), any());

            // When
            underTest.acceptSubmission(eventPayload);

            // Then
            verify(caseServiceMock, times(1)).sendEvent(submission.getCaseId(), eventJWT);
        }

        @Test
        void testRejectSubmission() throws ParseException, IOException {

            // Given
            final var submission = new Submission();
            submission.setSubmissionId(UUID.randomUUID());
            submission.setDestinationId(UUID.randomUUID());
            submission.setCaseId(UUID.randomUUID());

            final var problem = new AttachmentsMismatch();

            final EventPayload eventPayload = EventPayload.forRejectEvent(submission, List.of(problem));

            final var eventJWT = getResourceAsString("/signed_jwt.txt");
            when(securityEventServiceMock.createRejectSubmissionEvent(any())).thenReturn(SignedJWT.parse(eventJWT));
            when(validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
            doNothing().when(caseServiceMock).sendEvent(any(), any());

            // When
            underTest.rejectSubmission(eventPayload);

            // Then
            verify(caseServiceMock, times(1)).sendEvent(submission.getCaseId(), eventJWT);
        }

        @Test
        void testReceiveEventLogWithValidValidation() {

            // Given
            final var caseId = UUID.randomUUID();
            final var destinationId = UUID.randomUUID();

            final EventLogEntry eventLogEntry = EventLogEntry.builder()
                    .caseId(caseId)
                    .build();

            when(caseServiceMock.getEventLog(any(), any())).thenReturn(List.of(eventLogEntry));

            // When
            final List<EventLogEntry> eventLog = underTest.getEventLog(destinationId, caseId);

            // Then
            assertFalse(eventLog.isEmpty());
            assertThat(eventLog.get(0).getCaseId(), is(caseId));
        }

        @Test
        void testReceiveEventLogWithInvalidValidation() {

            // Given
            final var caseId = UUID.randomUUID();
            final var destinationId = UUID.randomUUID();

            when(caseServiceMock.getEventLog(any(), any())).thenThrow(new EventLogException("Invalid log"));

            // Then
            assertThrows(EventLogException.class, () -> underTest.getEventLog(destinationId, caseId));
        }

        @Test
        void testAttachmentValidationFailed() {

            // Given
            final Hash hash = new Hash();
            hash.setContent("hashValue");

            final ApiAttachment attachmentMetadata = new ApiAttachment();
            attachmentMetadata.setAttachmentId(UUID.randomUUID());
            attachmentMetadata.setHash(hash);

            final AttachmentForValidation attachmentForValidation = new AttachmentForValidation(attachmentMetadata, null, null);

            when(validationServiceMock.validateAttachments(anyList(), any())).thenReturn(ValidationResult.problem(new AttachmentsMismatch()));

            // When
            final ValidationResult validationResult = underTest.validateAttachments(List.of(attachmentForValidation), new AuthenticationTags());

            // Then
            assertTrue(validationResult.hasProblems());
            assertThat(validationResult.getProblems().get(0), instanceOf(AttachmentsMismatch.class));
        }

    }


    @Nested
    class ReplyRelatedTests {

        @Test
        void testListCases() {

            // Given
            final Case aCase = new Case();
            final Cases expectedCases = new Cases();

            expectedCases.setCases(Set.of(aCase));
            expectedCases.setCount(1);

            when(caseServiceMock.listCases(10, 0)).thenReturn(expectedCases);

            // When
            final Cases cases = underTest.listCases(10, 0);

            // Then
            assertThat(cases.getCount(), is(1));
            assertThat(cases.getCases().iterator().next(), is(aCase));
        }

        @Test
        void testGetSingleCase() {

            // Given
            final var caseId = UUID.randomUUID();

            final var expectedCase = new Case();
            expectedCase.setCaseId(caseId);

            when(caseServiceMock.getCase(any())).thenReturn(expectedCase);

            // When
            final Case aCase = underTest.getCase(caseId);

            // Then
            assertNotNull(aCase);
            assertThat(aCase, is(expectedCase));
        }

        @Test
        void testValidReplyMetadata() {

            // Given
            final Reply reply = new Reply();
            reply.setCaseId(UUID.randomUUID());

            final Case aCase = new Case();
            aCase.setDestinationId(UUID.randomUUID());

            when(caseServiceMock.getCase(any())).thenReturn(aCase);
            when(submissionServiceMock.getDestination(any())).thenReturn(new Destination());
            when(validationServiceMock.validateReplyMetadata(any(), any(), any(), any())).thenReturn(ValidationResult.ok());

            // When
            final ValidationResult validationResult = underTest.validateReplyMetadata(new Metadata(), reply, new AuthenticationTags());

            // Then
            assertTrue(validationResult.isValid());
        }

        @Test
        void testInvalidReplyMetadata() {

            // Given
            final Reply reply = new Reply();
            reply.setCaseId(UUID.randomUUID());

            final Case aCase = new Case();
            aCase.setDestinationId(UUID.randomUUID());

            when(caseServiceMock.getCase(any())).thenReturn(aCase);
            when(submissionServiceMock.getDestination(any())).thenReturn(new Destination());
            when(validationServiceMock.validateReplyMetadata(any(), any(), any(), any())).thenReturn(ValidationResult.error(new Exception("invalid metadata")));

            // When
            final ValidationResult validationResult = underTest.validateReplyMetadata(new Metadata(), new Reply(), new AuthenticationTags());

            // Then
            assertTrue(validationResult.hasError());
        }

        @Test
        void testGetReplyAuthTags() {

            // Given
            final Reply reply = new Reply();

            final AuthenticationTags expectedAuthTags = new AuthenticationTags();
            expectedAuthTags.setData("dataAuthTags");
            expectedAuthTags.setMetadata("metadataAuthTags");
            expectedAuthTags.setAttachments(Map.of(UUID.randomUUID(), "attachmentAuthTags"));

            when(caseServiceMock.getAuthenticationTags(reply)).thenReturn(expectedAuthTags);

            // When
            final AuthenticationTags authenticationTags = underTest.getReplyAuthenticationTags(reply);

            // Then
            assertThat(authenticationTags.getData(), is(expectedAuthTags.getData()));
            assertThat(authenticationTags.getMetadata(), is(expectedAuthTags.getMetadata()));
            assertThat(authenticationTags.getAttachments(), is(expectedAuthTags.getAttachments()));
        }

        @Test
        void testGetAvailableReplies() {

            // Given
            final ReplyForPickup replyForPickup = new ReplyForPickup();

            final RepliesForPickup expectedReplies = new RepliesForPickup();
            expectedReplies.setReplies(Set.of(replyForPickup));
            expectedReplies.setCount(1);

            when(replyServiceMock.getAvailableReplies(10, 0)).thenReturn(expectedReplies);

            // When
            final RepliesForPickup repliesForPickup = underTest.getAvailableReplies(10, 0);

            // Then
            assertThat(repliesForPickup.getCount(), is(1));
            assertThat(repliesForPickup.getReplies().iterator().next(), is(replyForPickup));
        }

        @Test
        void testGetSingleReply() {

            // Given
            final var replyId = UUID.randomUUID();

            final Reply expectedReply = new Reply();
            expectedReply.setReplyId(replyId);

            when(replyServiceMock.getReply(replyId)).thenReturn(expectedReply);

            // When
            final Reply reply = underTest.getReply(replyId);

            // Then
            assertNotNull(reply);
            assertThat(reply, is(expectedReply));
        }

        @Test
        void testReplyAttachmentUpload() {

            // Given
            final var replyId = UUID.randomUUID();
            final var attachmentId = UUID.randomUUID();
            final var encryptedAttachment = "encryptedAttachmentBytes";

            doNothing().when(replyServiceMock).uploadAttachment(any(), any(), any());

            // When
            underTest.uploadReplyAttachment(replyId, attachmentId, encryptedAttachment);

            // Then
            verify(replyServiceMock, times(1)).uploadAttachment(replyId, attachmentId, encryptedAttachment);
        }

        @Test
        void testGetReplyAttachment() {

            // Given
            final var replyId = UUID.randomUUID();
            final var attachmentId = UUID.randomUUID();
            final var expectedAttachment = "encryptedAttachmentContent";

            when(replyServiceMock.getAttachment(any(), any())).thenReturn(expectedAttachment);

            // When
            final String encryptedAttachment = underTest.getReplyAttachment(replyId, attachmentId);

            // Then
            assertThat(encryptedAttachment, equalTo(expectedAttachment));
        }

        @Test
        void testAnnounceReply() {

            // Given
            final var caseId = UUID.randomUUID();
            final List<UUID> announcedAttachments = List.of(UUID.randomUUID());
            final var announceReply = new AnnounceReply(caseId, announcedAttachments);

            final var announcedReply = new CreatedReply();
            announcedReply.setCaseId(caseId);
            announcedReply.setReplyId(UUID.randomUUID());

            when(replyServiceMock.announceReply(any())).thenReturn(announcedReply);

            // When
            final CreatedReply createdReply = underTest.announceReply(announceReply);

            // Then
            assertThat(createdReply, equalTo(announcedReply));
        }

        @Test
        void testSubmitReply() {

            // Given
            final var replyId = UUID.randomUUID();

            final SubmitReply submitReply = new SubmitReply();
            submitReply.setEncryptedData("encryptedData");
            submitReply.setEncryptedMetadata("encryptedMetadata");

            final SentReply expectedReply = new SentReply(replyId, UUID.randomUUID(), List.of());

            when(replyServiceMock.submitReply(any(), any())).thenReturn(expectedReply);

            // When
            final SentReply sentReply = underTest.submitReply(replyId, submitReply);

            // Then
            assertThat(sentReply, equalTo(expectedReply));
        }

        @Test
        void testAcceptReply() {

            // Given
            final var replyId = UUID.randomUUID();
            final var acceptReply = new AcceptReply(List.of(), new AuthenticationTags());
            final var expectedEvent = "dummy-accept-event";

            when(replyServiceMock.acceptReply(any(), any())).thenReturn(expectedEvent);

            // When
            final String acceptEvent = underTest.acceptReply(replyId, acceptReply);

            // Then
            assertThat(acceptEvent, equalTo(expectedEvent));
        }

        @Test
        void testRejectReply() {

            // Given
            final var replyId = UUID.randomUUID();
            final var expectedEvent = "dummy-reject-event";

            when(replyServiceMock.rejectReply(any(), any())).thenReturn(expectedEvent);

            // When
            final String rejectEvent = underTest.rejectReply(replyId, List.of(new TechnicalError()));

            // Then
            assertThat(rejectEvent, equalTo(expectedEvent));
        }
    }

    private String getResourceAsString(final String filename) throws IOException {
        return new String(FitConnectDefaultServiceTest.class.getResourceAsStream(filename).readAllBytes());
    }
}