package dev.fitko.fitconnect.core.crypto;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWEObject;
import com.nimbusds.jose.crypto.RSADecrypter;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.text.ParseException;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;

public class JWECryptoServiceTest {

    final CryptoService underTest;

    final RSAKey privateKey;

    public JWECryptoServiceTest() throws JOSEException {
        underTest = new JWECryptoService(new HashService());
        privateKey = generateRsaKey(2048);
    }

    @Test
    void encryptBytes() {
        // Given
        final RSAKey publicKey = privateKey.toPublicJWK();

        // When
        final String encryptedData = underTest.encryptBytes(publicKey, "some foo".getBytes(StandardCharsets.UTF_8), "text/plain");

        // Then
        assertNotNull(encryptedData);
        assertNotEquals("some foo", encryptedData);
    }

    @Test
    void encryptObject() throws IOException {
        // Given
        final RSAKey publicKey = privateKey.toPublicJWK();
        final Metadata metadata = new Metadata();

        // When
        final String encryptedData = underTest.encryptObject(publicKey, metadata, "application/json");

        // Then
        assertNotNull(encryptedData);
        assertThat(new ObjectMapper().readValue(underTest.decryptToBytes(privateKey, encryptedData), Metadata.class), is(metadata));
    }


    @Test
    void decryptBytes() {

        // Given
        final RSAKey publicKey = privateKey.toPublicJWK();
        final String dataFromSender = underTest.encryptBytes(publicKey, "some foo".getBytes(), "text/plain");

        // When
        final byte[] decrypted = underTest.decryptToBytes(privateKey, dataFromSender);

        // Then
        assertNotNull(decrypted);
        assertArrayEquals(decrypted, "some foo".getBytes(StandardCharsets.UTF_8));
    }


    @Test
    void hashBytes() {
        final var expectedHash = "6cb40f8c9a1f4411c6b83fefdd950009ec2906e3747e9146e7704234cdb6a7b11eefb9e6221fdad3cda7c5ece6a62ef535ca2ac84ba61226c9835062cf84db56";
        final byte[] contentToHash = "content to hash".getBytes(StandardCharsets.UTF_8);
        final String sha512Hash = underTest.hashBytes(contentToHash);

        assertEquals(expectedHash, sha512Hash);
        assertEquals(underTest.hashBytes(contentToHash), underTest.hashBytes(contentToHash));
    }

    @Test
    void testContentTypeIsSetInHeader() throws ParseException, JOSEException {

        // Given
        final RSAKey publicKey = privateKey.toPublicJWK();
        final String encryptedJWE = underTest.encryptBytes(publicKey, "some data".getBytes(), "text/plain");

        // When
        final JWEObject jwe = JWEObject.parse(encryptedJWE);
        jwe.decrypt(new RSADecrypter(privateKey));

        // Then
        assertNotNull(jwe);
        assertThat(jwe.getHeader().getContentType(), is("text/plain"));

    }

    public static RSAKey generateRsaKey(final int size) throws JOSEException {
        return new RSAKeyGenerator(size)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();
    }
}