package dev.fitko.fitconnect.sdk.examples;

import com.nimbusds.jose.jwk.JWK;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.crypto.JWKPair;
import dev.fitko.fitconnect.api.domain.model.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.event.Status;
import dev.fitko.fitconnect.api.domain.model.reply.RepliesForPickup;
import dev.fitko.fitconnect.api.domain.model.reply.SentReply;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.ReplyChannel;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.sender.ReceivedReply;
import dev.fitko.fitconnect.api.domain.sender.SendableSubmission;
import dev.fitko.fitconnect.api.domain.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.api.domain.subscriber.SendableReply;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.bootstrap.ApplicationConfigLoader;
import dev.fitko.fitconnect.client.bootstrap.ClientFactory;
import dev.fitko.fitconnect.client.util.ReplyChannelKeyGenerator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.URI;
import java.nio.file.Path;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.UUID;
import java.util.stream.Collectors;

public class ReplyExample {

    private static final Logger LOGGER = LoggerFactory.getLogger(ReplyExample.class);

    /**
     * Fill out config.yml in /resources and provide a DESTINATION_ID that matches the configuration (serviceType, data schema, etc.)
     */
    public static void main(String[] args) {

        // #######################
        // Setup
        // #######################

        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(Path.of("examples/src/main/resources/config.yml"));

        final SenderClient senderClient = ClientFactory.createSenderClient(config);
        final SubscriberClient subscriberClient = ClientFactory.createSubscriberClient(config);

        // #######################
        // Onlinedienst
        // #######################

        // generate reply channel encryption and decryption keys
        final JWKPair jwkPair = ReplyChannelKeyGenerator.generateKeyPair();

        // make sure to store the private key in a safe place
        final JWK privateKey = jwkPair.getPrivateKey();
        final JWK publicKey = jwkPair.getPublicKey();

        // initial submission with new case and FIT-Connect reply channel
        final SendableSubmission submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString("YOUR_DESTINATION_ID"))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test-Service")
                .setJsonData(getDataAsString("/data.json"), URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .addAttachment(Attachment.fromPath(Path.of("examples/src/main/resources/attachment.txt"), "application/text"))
                .setReplyChannel(ReplyChannel.ofFitConnect(publicKey, List.of("https://test.fitko.de")))
                .build();

        final SentSubmission sentSubmission = senderClient.send(submission);

        final Status sentSubmissionStatus = senderClient.getSubmissionStatus(sentSubmission);
        LOGGER.info("Submission status {}", sentSubmissionStatus.getState());

        // #######################
        // Fachbehörde
        // #######################

        // get all submission for case
        final List<ReceivedSubmission> receivedSubmissions = subscriberClient.getAvailableSubmissionsForDestination(UUID.fromString("YOUR_DESTINATION_ID"))
                .stream()
                .filter(s -> s.getCaseId().equals(sentSubmission.getCaseId()))
                .map(subscriberClient::requestSubmission)
                .collect(Collectors.toList());

        // make sure to get the latest submission from case with most current reply channel public key
        // see https://docs.fitko.de/fit-connect/docs/sdks/java-sdk/receiver#empf%C3%A4nger-sendet-antwort-%C3%BCber-fit-connect
        receivedSubmissions.sort(Collections.reverseOrder());
        final ReceivedSubmission latestSubmission = receivedSubmissions.get(0);

        final SendableReply sendableReply = SendableReply.from(latestSubmission)
                .setJsonData(getDataAsString("/data.json"), URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .addAttachment(Attachment.fromPath(Path.of("examples/src/main/resources/attachment.txt"), "application/text"))
                .build();

        // send reply to online service
        final SentReply sentReply = subscriberClient.sendReply(sendableReply);
        final Status replyStatus = subscriberClient.getReplyStatus(sentReply);
        LOGGER.info("Reply status {}", replyStatus.getState());

        // #######################
        // Onlinedienst
        // #######################

        // receive available replies
        final RepliesForPickup repliesForPickup = senderClient.getAvailableReplies();
        LOGGER.info("{} new replies are available", repliesForPickup.getCount());

        // // receive single reply and accept
        final ReceivedReply receivedReply = senderClient.getReply(sentReply.getReplyId(), privateKey);
        LOGGER.info("Received reply data {}", receivedReply.getDataAsString());

        // accept correct reply
        receivedReply.acceptReply();
    }

    private static String getDataAsString(final String path) {
        try {
            return new String(Objects.requireNonNull(SenderExample.class.getResourceAsStream(path)).readAllBytes());
        } catch (final IOException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }
}
